package com.example.schedulerapp;

import java.io.*;
import java.net.*;
import java.time.LocalTime;

/*
    Class Name: ReceiveThread
    Description: This class handles messages sent from the server, and the required parsing to be used for in the Model class.
 */
public class ReceiveThread extends Thread {
    private BufferedReader reader;
    private final Model client;

    /*
    Name: ReceiveThread
    Parameters:
        Socket socket: socket for communicating to the server.
        Model client: class that contains the employee and shift hashmaps.
    Description: Constructor class that gets the input stream to be read when server messages are sent.
    Return: ReceiveThread
     */
    public ReceiveThread(Socket socket, Model client) {
        this.client = client;

        try {
            InputStream input = socket.getInputStream();
            reader = new BufferedReader(new InputStreamReader(input));
        } catch (IOException exception) {
            System.out.println("Error getting input stream: " + exception.getMessage());
            exception.printStackTrace();
        }
    }

    /*
    Name: Run
    Description: Responsible for handling and parsing messages from the server. This is also
    where the receive-thread begins.
     */
    public void run() {
        while(true) {
            try {
                String response = reader.readLine(); // Waits for messages from the server.
                String[] args = response.split("/");
                switch (args[0]) {
                    case "allEmployees" -> allEmployees(args);
                    case "allShifts" -> allShifts(args);
                    case "addEmployee" -> addEmployee(args[1]);
                    case "editEmployee" -> editEmployee(args[1]);
                    case "removeEmployee" -> removeEmployee(args[1]);
                    case "addShift" -> addShift(args[1]);
                    case "removeShift" -> removeShiftByID(args[1]);
                    case "editShift" -> editShift(args[1]);
                    case "editShiftAvailability" -> editShiftAvailability(args[1], args[2]);
                    case "newEmployeeAvailability" -> addAvailability(args[1]);
                    case "allAvailabilities" -> allAvailability(args);
                    case "editAvailability" -> editAvailability(args[1]);
                    case "addTimeOff" -> addTimeOff(args[1]);
                    case "removeTimeOff" -> removeTimeOff(args[1]);
                    case "setTimeOffApproval" -> setTimeOffApproval(args[1]);
                    case "allTimeOff" -> allTimeOff(args);
                    case "allPositions" -> allPositions(args);
                    case "addPosition" -> addPosition(args[1], args[2]);
                    case "editPosition" -> editPosition(args[1], args[2]);
                    case "removePosition" -> removePosition(args[1]);
                    case "addEmployeePosition" -> addEmployeePosition(args[1], args[2]);
                    case "removeEmployeePosition" -> removeEmployeePosition(args[1], args[2]);
                    case "removePositionAllEmployees" -> removePositionAllEmployees(args[1]);
                    case "checkedIn" -> checkIn(args[1]);
                }

            } catch (IOException exception) {
                System.out.println("Dropped connection from server: " + exception.getMessage());
                exception.printStackTrace();
                break;
            }

        }
    }

    /*
    Name: allEmployees
    Parameters:
        String[] allEmployees: All employee information sent from server. Employee fields are is separated by a '.'
        delimiter.
    Description: Separates all employees and adds them to the local employee HashMap by their ID.
    Return: void
     */
    private void allEmployees(String[] allEmployees) {
        for (String employee : allEmployees) {
            if (!employee.equals("allEmployees")){
               addEmployee(employee);
            }
        }
    }

    /*
    Name: addEmployee
    Parameters:
        String employeeData: All the information for a given employee separated by ','
    Description: Adds an employee to the local employee HashMap by their ID.
    Return: void
     */
    private void addEmployee(String employeeData) {
        try {
            String[] dataSplit = employeeData.split(",");
            this.client.employees.put(Integer.parseInt(dataSplit[0]), new Employee(employeeData));
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: editEmployee
    Parameters:
        String employeeData: All the information for a given employee separated by ','
    Description: Edits an employee on the local employee HashMap.
    Return: void
     */
    private void editEmployee(String employeeData) {
        try {
            String[] dataSplit = employeeData.split(",");
            this.client.employees.remove(Integer.parseInt(dataSplit[0]));
            addEmployee(employeeData);
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: RemoveEmployee
    Parameters:
        String employeeID: The employeeID of the employee.
    Description: Removes an employee from the employee HashMap by there employee ID.
    Return: void
     */
    private void removeEmployee(String employeeID) {
        try {
            int intID = Integer.parseInt(employeeID);
            this.client.employees.remove(intID);
            for (Shift shift : this.client.shifts.values()) {
                if (shift.getEmployeeID() == intID) {
                    this.client.shifts.remove(shift.getShiftID());
                }
            }
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: allShifts
    Parameters:
        String[] allShifts: All the shifts to be added locally from the server in the form String[].
    Description: Adds all the shifts to the local shift hashmap by the shift id.
    Return: void
     */
    private void allShifts(String[] allShifts) {
        for (String shift : allShifts) {
            if (!shift.equals("allShifts")){
                addShift(shift);
            }
        }
    }

    /*
    Name: addShift
    Parameters:
        String shiftData: a string containing shift data from the server
    Description: Adds a shift to the local shift Hashmap by shift ID.
    Return: void
     */
    private void addShift(String shiftData) {
        try {
            String[] dataSplit = shiftData.split(",");
            this.client.shifts.put(Integer.parseInt(dataSplit[4]), new Shift(shiftData));
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: removeShiftByID
    Parameters:
        String id: The shift id of an existing shift.
    Description: Removes a shift from the local shift HashMap.
    Return: void
     */
    private void removeShiftByID(String id) {
        try {
            this.client.shifts.remove(Integer.parseInt(id));
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: editShift
    Parameters:
        String editedShift: The information of the editedShift from the server.
    Description: Edits an existing shift.
    Return: void
     */
    private void editShift(String editedShift) {
        String[] shiftSplit = editedShift.split(",");
        try {
            Integer.parseInt(shiftSplit[4]);
            removeShiftByID(shiftSplit[4]);
            addShift(editedShift);
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: editShiftAvailability
    Parameters:
        String shiftID: The id of the shift
        String available: "true" if the shift is available, and "false" if its not.
    Description: Changes the availability of a shift, so we know if it's been "up-for-grabs" by another employee.
    Return: void
     */
    private void editShiftAvailability(String shiftID, String available) {
        try {
            this.client.shifts.get(Integer.parseInt(shiftID)).setAvailable(Boolean.parseBoolean(available));
        } catch (NumberFormatException exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: addAvailability
    Parameters:
        String availabilityData: An employee's availability where each detail is separated by commas.
    Description: Converts and adds the employee's availability data into there employee object.
    Return: void
     */
    private void addAvailability(String availabilityData) {
        try {
            String[] availabilitySplit = availabilityData.split(",");
            client.employees.get(Integer.parseInt(availabilitySplit[0])).populateAvailability(availabilitySplit);
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: allAvailabilities
    Parameters:
        String[] allAvailabilities: All availabilities where each employee's availability is separated by slashes and
        each detail is separated by commas.
    Description: Converts and adds all the availabilities for each employee.
    Return: void
     */
    private void allAvailability(String[] allAvailabilities) {
        for (String availability : allAvailabilities) {
            if (!availability.equals("allAvailabilities")) {
                addAvailability(availability);
            }
        }
    }

    /*
    Name: editAvailabilities
    Parameters:
        String editAvailabilities: The edited availability for a specific day of the week.
    Description: Changes the availability for a specific day of the week for an employee.
    Return: void
     */
    private void editAvailability(String editAvailability) {
        String[] editSplit = editAvailability.split(",");
        int employeeID = Integer.parseInt(editSplit[0]);
        int dayOfTheWeek = Integer.parseInt(editSplit[1]);
        LocalTime startTime = LocalTime.parse(editSplit[2]);
        LocalTime endTime = LocalTime.parse(editSplit[3]);
        client.getEmployee(employeeID).updateAvailability(dayOfTheWeek, startTime, endTime);
    }

    /*
    Name: addTimeOff
    Parameters:
        String timeOffData: The time off data separated by commas.
    Description: Adds the time off data to a time off object and stores it in a hashmap by its id.
    The time off id is also stored in the employees time off list.
    Return: void
     */
    private void addTimeOff(String timeOffData) {
        String[] timeOffSplit = timeOffData.split(",");
        int timeOffID = Integer.parseInt(timeOffSplit[0]);
        client.timeOff.put(timeOffID, new TimeOff(timeOffSplit));
        client.employees.get(Integer.parseInt(timeOffSplit[1])).addTimeOffID(timeOffID);
    }

    /*
    Name: removeTimeOff
    Parameters:
        String timeOffID: The id of the time off request.
    Description: Removes the time off request from the local time off hashmap.
    Return: void
     */
    private void removeTimeOff(String timeOffID) {
        try {
            int id = Integer.parseInt(timeOffID);
            int employeeID = this.client.timeOff.get(id).getEmployeeID();
            client.getEmployee(employeeID).removeTimeOffID(id);
            client.timeOff.remove(id);
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: setTimeOffApproval
    Parameters:
        String approvalData: "true" if the time off was approved and "false" if it wasn't
    Description: Changes the approval status of a request if been accepted by a manager.
    Return: void
     */
    private void setTimeOffApproval(String approvalData) {
        try {
            String[] dataSplit = approvalData.split(",");
            int id = Integer.parseInt(dataSplit[0]);
            this.client.timeOff.get(id).setApproval(Boolean.getBoolean(dataSplit[1]));
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: allTimeOff
    Parameters:
        String[] allTimeOff: All the time off requests where each request is separated by slashes and each detail is
        separated by commas.
    Description: Separates all the time off requests as TimeOff objects and stored in the time off hash map keyed to
    their special id.
    Return: void
     */
    private void allTimeOff(String[] allTimeOff) {
        for (String timeOff : allTimeOff) {
            if (!timeOff.equals("allTimeOff")) {
                addTimeOff(timeOff);
            }
        }
    }

    /*
    Name: addEmployeePosition
    Parameters:
        String employeeID: The id of the employee.
        String position: The position to be added.
    Description: Adds the position to the position array list in the employee object.
    Return: void
     */
    private void addEmployeePosition(String employeeID, String position) {
        try {
            int id = Integer.parseInt(employeeID);
            this.client.getEmployee(id).addPosition(position);
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: removeEmployeePosition
    Parameters:
        String employeeID: The id of the employee.
        String position: The position to be added.
    Description: Removes the position to in the position array list in the employee object.
    Return: void
     */
    private void removeEmployeePosition(String employeeID, String position) {
        try {
            int id = Integer.parseInt(employeeID);
            this.client.getEmployee(id).removePosition(position);
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: addPosition
    Parameters:
        String position: The position to be added.
        String wage: The wage of the position
    Description: Adds a position positions hashmap where the position is the key and value is the wage.
    Return: void
     */
    private void addPosition(String position, String wage) {
        try {
            this.client.positions.put(position, Float.parseFloat(wage));
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: editPosition
    Parameters:
        String position: The name of an existing position.
        String wage: The new wage.
    Description: Changes the wage of a position.
    Return: void
     */
    private void editPosition(String position, String wage) {
        try {
            this.client.positions.put(position, Float.parseFloat(wage));
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /*
    Name: removePosition
    Parameters:
        String position: The name of an existing position.
    Description: Removes a position for the list of positions.
    Return: void
     */
    private void removePosition(String position) {
        this.client.positions.remove(position);
    }

    /*
    Name: allPositions
    Parameters:
        String[] positionData: All the position data where each position is separated by slashes and each detail is
        separated by commas.
    Description: Adds all positions to the local positions hashmap where the position is the key and the value is the
    wage.
    Return: void
     */
    private void allPositions(String[] positionData){
        for (String position : positionData) {
            if (!position.equals("allPositions")) {
                addPosition(position.split(",")[0], position.split(",")[1]);
            }
        }
    }

    /*
    Name: removePositionAllEmployees
    Parameters:
        String position: The name of an existing position.
    Description: Removes a position from all employees positions array list.
    Return: void
     */
    private void removePositionAllEmployees(String position) {
        for (int id : this.client.employees.keySet()) {
            removeEmployeePosition(Integer.toString(id), position);
        }
    }

    /*
    Name: checkIn
    Parameters:
        String shiftID:
    Description: Marks a shift as checked in, essentially meaning it's been worked.
    Return: void
     */
    private void checkIn(String shiftID) {
        try {
            this.client.shifts.get(Integer.parseInt(shiftID)).setCheckedIn();
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

}
